@@ -4,7 +4,6 @@ from __future__ import absolute_import, unicode_literals
 
 import logging
 import socket
-
 from collections import defaultdict
 from warnings import warn
 
@@ -12,10 +11,9 @@ from vine import ensure_promise
 
 from . import spec
 from .abstract_channel import AbstractChannel
-from .exceptions import (
-    ChannelError, ConsumerCancelled,
-    RecoverableChannelError, RecoverableConnectionError, error_for_code,
-)
+from .exceptions import (ChannelError, ConsumerCancelled,
+                         RecoverableChannelError, RecoverableConnectionError,
+                         error_for_code, MessageNacked)
 from .five import Queue
 from .protocol import queue_declare_ok_t
 
@@ -95,6 +93,7 @@ class Channel(AbstractChannel):
         spec.method(spec.Tx.SelectOk),
         spec.method(spec.Confirm.SelectOk),
         spec.method(spec.Basic.Ack, 'Lb'),
+        spec.method(spec.Basic.Nack, 'Lb'),
     }
     _METHODS = {m.method_sig: m for m in _METHODS}
 
@@ -140,6 +139,7 @@ class Channel(AbstractChannel):
             spec.Basic.Deliver: self._on_basic_deliver,
             spec.Basic.Return: self._on_basic_return,
             spec.Basic.Ack: self._on_basic_ack,
+            spec.Basic.Nack: self._on_basic_nack,
         })
 
     def collect(self):
@@ -219,12 +219,14 @@ class Channel(AbstractChannel):
             if is_closed:
                 return
 
+            self.is_closing = True
             return self.send_method(
                 spec.Channel.Close, argsig,
                 (reply_code, reply_text, method_sig[0], method_sig[1]),
                 wait=spec.Channel.CloseOk,
             )
         finally:
+            self.is_closing = False
             self.connection = None
 
     def _on_close(self, reply_code, reply_text, class_id, method_id):
@@ -274,10 +276,11 @@ class Channel(AbstractChannel):
                 is the ID of the method.
         """
         self.send_method(spec.Channel.CloseOk)
-        self._do_revive()
-        raise error_for_code(
-            reply_code, reply_text, (class_id, method_id), ChannelError,
-        )
+        if not self.connection.is_closing:
+            self._do_revive()
+            raise error_for_code(
+                reply_code, reply_text, (class_id, method_id), ChannelError,
+            )
 
     def _on_close_ok(self):
         """Confirm a channel close.
@@ -1137,8 +1140,8 @@ class Channel(AbstractChannel):
                 True.
 
         Returns a tuple containing 3 items:
-            the name of the queue (essential for automatically-named queues)
-            message count
+            the name of the queue (essential for automatically-named queues),
+            message count and
             consumer count
         """
         self.send_method(
@@ -1217,6 +1220,8 @@ class Channel(AbstractChannel):
                 client should not wait for a reply method.  If the
                 server could not complete the method it will raise a
                 channel or connection exception.
+
+        If nowait is False, returns the number of deleted messages.
         """
         return self.send_method(
             spec.Queue.Delete, argsig,
@@ -1277,7 +1282,7 @@ class Channel(AbstractChannel):
                 server could not complete the method it will raise a
                 channel or connection exception.
 
-        if nowait is False, returns a message_count
+        If nowait is False, returns a number of purged messages.
         """
         return self.send_method(
             spec.Queue.Purge, argsig, (0, queue, nowait),
@@ -1559,14 +1564,26 @@ class Channel(AbstractChannel):
         """
         p = self.send_method(
             spec.Basic.Consume, argsig,
-            (0, queue, consumer_tag, no_local, no_ack, exclusive,
-             nowait, arguments),
+            (
+                0, queue, consumer_tag, no_local, no_ack, exclusive,
+                nowait, arguments
+            ),
             wait=None if nowait else spec.Basic.ConsumeOk,
+            returns_tuple=True
         )
 
-        # XXX Fix this hack
-        if not nowait and not consumer_tag:
-            consumer_tag = p
+        if not nowait:
+            # send_method() returns (consumer_tag,) tuple.
+            # consumer_tag is returned by broker using following rules:
+            # * consumer_tag is not specified by client, random one
+            #   is generated by Broker
+            # * consumer_tag is provided by client, the same one
+            #   is returned by broker
+            consumer_tag = p[0]
+        elif nowait and not consumer_tag:
+            raise ValueError(
+                'Consumer tag must be specified when nowait is True'
+            )
 
         self.callbacks[consumer_tag] = callback
 
@@ -1632,7 +1649,8 @@ class Channel(AbstractChannel):
                 reliability.  Messages can get lost if a client dies
                 before it can deliver them to the application.
 
-        Non-blocking, returns a message object, or None.
+        Non-blocking, returns a amqp.basic_message.Message object,
+        or None if queue is empty.
         """
         ret = self.send_method(
             spec.Basic.Get, argsig, (0, queue, no_ack),
@@ -1667,6 +1685,11 @@ class Channel(AbstractChannel):
         configuration and distributed to any active consumers when the
         transaction, if any, is committed.
 
+        When channel is in confirm mode (when Connection parameter
+        confirm_publish is set to True), each message is confirmed. When
+        broker rejects published message (e.g. due internal broker
+        constrains), MessageNacked exception is raised.
+
         PARAMETERS:
             exchange: shortstr
 
@@ -1727,6 +1750,15 @@ class Channel(AbstractChannel):
         if not self.connection:
             raise RecoverableConnectionError(
                 'basic_publish: connection closed')
+
+        client_properties = self.connection.client_properties
+        if client_properties['capabilities']['connection.blocked']:
+            try:
+                # Check if an event was sent, such as the out of memory message
+                self.connection.drain_events(timeout=0)
+            except socket.timeout:
+                pass
+
         try:
             with self.connection.transport.having_timeout(timeout):
                 return self.send_method(
@@ -1738,11 +1770,18 @@ class Channel(AbstractChannel):
     basic_publish = _basic_publish
 
     def basic_publish_confirm(self, *args, **kwargs):
+
+        def confirm_handler(method, *args):
+            # When RMQ nacks message we are raising MessageNacked exception
+            if method == spec.Basic.Nack:
+                raise MessageNacked()
+
         if not self._confirm_selected:
             self._confirm_selected = True
             self.confirm_select()
         ret = self._basic_publish(*args, **kwargs)
-        self.wait(spec.Basic.Ack)
+        # Waiting for confirmation of message.
+        self.wait([spec.Basic.Ack, spec.Basic.Nack], callback=confirm_handler)
         return ret
 
     def basic_qos(self, prefetch_size, prefetch_count, a_global,
@@ -2039,3 +2078,7 @@ class Channel(AbstractChannel):
     def _on_basic_ack(self, delivery_tag, multiple):
         for callback in self.events['basic_ack']:
             callback(delivery_tag, multiple)
+
+    def _on_basic_nack(self, delivery_tag, multiple):
+        for callback in self.events['basic_nack']:
+            callback(delivery_tag, multiple)
