@@ -13,7 +13,8 @@ class test_Message:
             channel=Mock(name='channel'),
             application_headers={'h': 'v'},
         )
-        m.delivery_info = {'delivery_tag': '1234'},
+        m.delivery_info = {'delivery_tag': '1234'}
         assert m.body == 'foo'
         assert m.channel
         assert m.headers == {'h': 'v'}
+        assert m.delivery_tag == '1234'
